Attribute VB_Name = "Perturbation"
Option Explicit
Dim NoOfTables As Integer, NoOfMaps As Integer, NoOfVersions As Integer

Dim SASPath As String, PerturbPath As String, PerturbOutputsPath As String
Sub main()
' PROGRAM TO READ IN RAW SAS/LBS DATA AND WRITE OUT:
'   (i)  RAW SAS/LBS DATA
'   (ii) PERTURBED VARIANTS OF (i)

'To turn on full error trail, search and replace ''print #temp with 'print #temp

'To increase number of perturbation variants, need to redimension all arrays currently with
'a 1 to 4 element to a 1 to NoOfVersions element

' file handles for current table
Dim filelist As Integer, threshlist As Integer, temp As Integer
Dim raw_input As Integer, mapfile As Integer, tablemaps As Integer
Dim thresh_out As Integer, perturb_1 As Integer
Dim perturb_2 As Integer, perturb_3 As Integer
Dim perturb_4 As Integer, perturb_5 As Integer, perturb_6 As Integer

Dim hholds, residents, ED, iteration As Integer

Dim rows As Integer, cols As Integer 'no. of rows and columns in each table
Dim map(1 To 40, 0 To 40) As Integer 'map of counts to marginals (1:40)
                                              'by mapping (1:40) by
                                              '-1 = marginal; >0 = contributor to marginal
                                              'table 0: 1= row marg; 2=col marg.
  
Dim contrib_cells(1 To 40, 1 To 40) As Integer 'no. of cells contributing to each marginal
                                                        '1:40 - x; 1:40 - y
Dim valid_contrib_cells(1 To 1600) As Integer 'flag for no. of cells (1:1600) on which marginals are based
Dim temp_table(1 To 40, 1 To 40, 0 To 6) As Integer 'ED x row x col x table version
  'v0 = original; v2 = rounded (base 5); v2 = rounded (base 3);
  'v3 = small nos [margins = sum of internal cells];
  'v4=barndardisation (p=0.1); v5=barndardisation (p=0.02);
  'v6 = small nos [all cells]
  
Dim zoneid As String * 8, tablename As String
Dim InconsistentMarginals As Integer

SASPath = "d:\research\disclosure control\disclosure_vb\SAS_Inputs\"
PerturbPath = "d:\research\disclosure control\disclosure_vb\Perturb_v3\"
PerturbOutputsPath = "d:\research\disclosure control\disclosure_vb\Perturb_v3\Outputs\"

'==== Open general input files ====

'Open file for storing run-time test output
temp = FreeFile
Open PerturbPath & "perturb_out.txt" For Output As #temp

'Open file for storing table mappings converted to matrix format [for checking purposes]
tablemaps = FreeFile
Open PerturbPath & "Table mapping check.txt" For Output As #tablemaps

'Open file listing tables to be processed
filelist = FreeFile
Open PerturbPath & "Perturbation input tables.txt" For Input As #filelist

NoOfTables = 0

'=== Perturb input tables one at a time ===
Do While Not EOF(filelist)
  
  Input #filelist, tablename, InconsistentMarginals
  Print #temp, " === " & Trim(tablename) & " - ";
  If (InconsistentMarginals = 1) Then
    Print #temp, "has inconsistent marginals requiring correction before perturbation"
  Else
    Print #temp, "has consistent marginals"
  End If
  
  NoOfTables = NoOfTables + 1

  'Open file containing ED threshold data [reopen for each table ==> rewinding!]
  threshlist = FreeFile
  Open SASPath & "threshold.fmt" For Input As #threshlist

  Call initialise

  ' Open table-specific input and output files
  Call open_table_specific_files(raw_input, mapfile, thresh_out, _
     perturb_1, perturb_2, perturb_3, perturb_4, perturb_5, perturb_6, tablename)

  ' Read in table-specific mappings
  Call table_mapping(tablename, mapfile, tablemaps, temp, rows, cols, _
    map, contrib_cells, valid_contrib_cells)


'Scroll through EDs one at a time (listed in threshold.fmt)

ED = 1
Do While Not EOF(threshlist)
'Do While ED < 4
  ED = ED + 1

  '(1) Read in no. of households and residents for current area

  Input #threshlist, zoneid, residents, hholds

  '(2) Read in current ED data for each table

   Call get_raw_table(raw_input, temp, rows, cols, temp_table, zoneid)
   
   'If SAS Table with known discrepancies between marginals and table cell counts,
   '(e.g. Age part of SAS Table 6) adjust marginals to fit cell counts
   If InconsistentMarginals = 1 Then Call Recalc_raw_marginals(temp_table, map, rows, cols, temp)
   
   
  '(3) If area >= threshold
  
    'If (hholds >= 40) And (residents >= 100) Then
    'Print #temp, "Valid ED: ", zoneid, hholds, residents

      '(3a) Perturb table counts (including marginals)
       Call perturb_counts(temp, rows, cols, temp_table, tablename)
       
      '(3b) Recalculate table marginals to fit perturbed counts
       Call Recalc_marginals(temp_table, map, rows, cols, temp, tablename)
                     
      '(3c) Write out original and perturbed versions of tables
       Call output_data(thresh_out, perturb_1, perturb_2, perturb_3, _
           perturb_4, perturb_5, perturb_6, rows, cols, temp_table, zoneid)

    'Else 'area below threshold [discard data; do not perturb]
    '  Print #temp, "INVALID ED: ", zoneid, hholds, residents
    'End If

  Loop 'next ED in list

  Close #threshlist

  Call close_table_specific_files(raw_input, mapfile, thresh_out, _
     perturb_1, perturb_2, perturb_3, perturb_4, perturb_5, perturb_6)
     
Loop 'next table in list

Close ' Close all remaining open files

End Sub
Sub initialise()

'ValidED = 0
'ValidEDs = 0
NoOfTables = 0
NoOfVersions = 4 'No. of table versions, including original + all perturbed versions

End Sub
Sub open_table_specific_files(raw_input, mapfile, thresh_out, _
  perturb_1, perturb_2, perturb_3, perturb_4, perturb_5, perturb_6, tablename)

'For current input table, open the following input and output files
' (i) raw_input - raw SAS/LBS data for table
' (ii) mapfile - row and col. marginal mapping for table
' (iii) thresh_out - SAS/LBS data for areas at or above threshold
' (iv) perturb_1 - Rounding to base 5
' (v) perturb_2 - Rounding to base 3
' (vi) perturb_3 - Small no. rounding
' (vii) perturb_4 - Barnardisation (p=0.1)
' (viii) perturb_5 - Barnardisation (p=0.02)

' (i) raw_input - raw SAS/LBS data for table
  raw_input = FreeFile
  Open SASPath & Trim(tablename) & "_raw.fmt" For Input As #raw_input

' (ii) mapfile - row and col. marginal mapping for table
  mapfile = FreeFile
  Open PerturbPath & Trim(tablename) & ".map" For Input As #mapfile

' (iii) threshold- SAS/LBS data for areas at or above threshold
  thresh_out = FreeFile
  Open PerturbOutputsPath & Trim(tablename) & "_v0.fmt" For Output As #thresh_out

' (iv) perturb_1
  perturb_1 = FreeFile
  'print #temp, Trim(tablename)
  Open PerturbOutputsPath & Trim(tablename) & "_v1.fmt" For Output As #perturb_1
  
' (v) perturb_2
  perturb_2 = FreeFile
  Open PerturbOutputsPath & Trim(tablename) & "_v2.fmt" For Output As #perturb_2
  
' (vi) perturb_3
  perturb_3 = FreeFile
  Open PerturbOutputsPath & Trim(tablename) & "_v3.fmt" For Output As #perturb_3
  
' (vi) perturb_4
  perturb_4 = FreeFile
  Open PerturbOutputsPath & Trim(tablename) & "_v4.fmt" For Output As #perturb_4
  
' (vi) perturb_5
  perturb_5 = FreeFile
  Open PerturbOutputsPath & Trim(tablename) & "_v5.fmt" For Output As #perturb_5
  
' (vii) perturb_6
  perturb_6 = FreeFile
  Open PerturbOutputsPath & Trim(tablename) & "_v6.fmt" For Output As #perturb_6
  
End Sub
Sub table_mapping(tablename, mapfile, tablemaps, temp, rows, cols, _
  map, contrib_cells, valid_contrib_cells)
Dim t, mapping, rowcol, x, y, marginal, origvalue As Integer

'Find dimensions of current table
Input #mapfile, rows, cols

'print #temp, "Table "; t; " rows "; rows(t); " cols "; cols(t)

'Set all table cells in the array contrib_cells to 1

For y = 1 To rows
  For x = 1 To cols
    contrib_cells(x, y) = 1
  Next x
Next y


mapping = 0

Do While Not EOF(mapfile)

   mapping = mapping + 1

  'Check to see if mapping is for row or column
  Input #mapfile, rowcol

  'Find no. of cols/rows wide table is, as appropriate

  If (rowcol = 1) Then
    map(mapping, 0) = 1
    x = cols 'Mapping is for row X-cols wide
  End If
  If (rowcol = 2) Then
    map(mapping, 0) = 2
    x = rows 'Mapping is for col X-rows wide
  End If

  'Read in mapping and identify marginal

  For y = 1 To x
    Input #mapfile, map(mapping, y)
    If (map(mapping, y) = -1) Then marginal = y
  Next y 'next row/col

  'print #temp, t; " "; rowcol; " ";
  'For y = 1 To 40
    'print #temp, map(mapping, y); " ";
  'Next y
  'print #temp,
  
  'Count up number of cells contributing to marginal
  
  If (map(mapping, 0) = 1) Then 'col. marginal
    For y = 1 To rows
      origvalue = contrib_cells(marginal, y)
      For x = 1 To cols
        If map(mapping, x) > 0 Then contrib_cells(marginal, y) = _
           contrib_cells(marginal, y) + contrib_cells(x, y)
      Next x 'next col.
      contrib_cells(marginal, y) = contrib_cells(marginal, y) - origvalue
    Next y 'next row
  End If
  
  If (map(mapping, 0) = 2) Then 'row marginal
    For x = 1 To cols
      origvalue = contrib_cells(x, marginal)
      For y = 1 To rows
        If map(mapping, y) > 0 Then contrib_cells(x, marginal) = _
          contrib_cells(x, marginal) + contrib_cells(x, y)
      Next y 'next col.
      contrib_cells(x, marginal) = contrib_cells(x, marginal) - origvalue
    Next x 'next row
  End If
  
'  Print #temp, "Count_cells"; " Table "; t; " Mapping "; mapping; " marginal ", marginal
'  For y = 1 To 40
'    For x = 1 To 40
'      Print #temp, f(contrib_cells(x, y), 5, 0);
'    Next x
'    Print #temp,
'  Next y
  
Loop 'next mapping for current table

NoOfMaps = mapping

'print #temp, "Table ", t, "no. of mappings ", mapping

  Print #tablemaps, " Table Mapping for "; tablename
  For y = 1 To rows
    For x = 1 To cols
      Print #tablemaps, f(contrib_cells(x, y), 5, 0);
    Next x
    Print #tablemaps,
  Next y

'Record all contributing-cell counts

For y = 1 To rows
  For x = 1 To cols
    valid_contrib_cells(contrib_cells(x, y)) = 1
  Next x
Next y

'For x = 1 To 1600
'  If valid_contrib_cells(x) > 0 Then
'    Print #temp, "Valid_contrib_cells "; x
'  End If
'Next x

End Sub

Sub get_raw_table(raw_input, temp, rows, cols, temp_table, zoneid)
Dim x, y As Integer

  Input #raw_input, zoneid
  
  For y = 1 To rows
     For x = 1 To cols
        Input #raw_input, temp_table(x, y, 0)
     Next x
  Next y

  'print #temp, zoneid
'  For y = 1 To rows
'     For x = 1 To cols
'        Print #temp, f(temp_table(x, y, 0), 5, 0);
'     Next x
'     Print #temp,
'  Next y

End Sub

Sub write_table(temp, version, rows, cols, temp_table, tablename)
Dim x, y As Integer

  Print #temp, "Table "; tablename; " version "; version
  
  For y = 1 To rows
     For x = 1 To cols
        Print #temp, "  "; Trim(temp_table(x, y, version));
     Next x
     Print #temp,
  Next y
  
End Sub
Sub output_data(thresh_out, perturb_1, perturb_2, perturb_3, perturb_4, perturb_5, _
   perturb_6, rows, cols, temp_table, zoneid)
Dim t As Integer, x As Integer, y As Integer
Dim table8_out As Integer

  Print #thresh_out, zoneid
    
  For y = 1 To rows
     For x = 1 To cols
        Print #thresh_out, "  "; Trim(temp_table(x, y, 0));
     Next x
     Print #thresh_out,
  Next y

  Print #perturb_1, zoneid
  For y = 1 To rows
     For x = 1 To cols
        Print #perturb_1, "  "; Trim(temp_table(x, y, 1));
     Next x
     Print #perturb_1,
  Next y

  Print #perturb_2, zoneid
  For y = 1 To rows
     For x = 1 To cols
        Print #perturb_2, "  "; Trim(temp_table(x, y, 2));
     Next x
     Print #perturb_2,
  Next y

  Print #perturb_3, zoneid
  For y = 1 To rows
     For x = 1 To cols
        Print #perturb_3, "  "; Trim(temp_table(x, y, 3));
     Next x
     Print #perturb_3,
  Next y

  Print #perturb_4, zoneid
  For y = 1 To rows
     For x = 1 To cols
        Print #perturb_4, "  "; Trim(temp_table(x, y, 4));
     Next x
     Print #perturb_4,
  Next y

  Print #perturb_5, zoneid
  For y = 1 To rows
     For x = 1 To cols
        Print #perturb_5, "  "; Trim(temp_table(x, y, 5));
     Next x
     Print #perturb_5,
  Next y

  Print #perturb_6, zoneid
  For y = 1 To rows
     For x = 1 To cols
        Print #perturb_6, "  "; Trim(temp_table(x, y, 6));
     Next x
     Print #perturb_6,
  Next y


End Sub
      
Sub perturb_counts(temp, rows, cols, temp_table, tablename)
Dim x, y, rmndr As Integer
Dim prob, prob1, prob2 As Double

For x = 1 To cols
  For y = 1 To rows
    
    ' Random rounding to nearest 5

    rmndr = temp_table(x, y, 0) Mod 5
    
    prob = Rnd
    If rmndr = 1 Then
      If prob <= (4# / 5#) Then
        rmndr = 0
      Else
        rmndr = 5
      End If
    ElseIf (rmndr = 2) Then
      If prob <= (3# / 5#) Then
        rmndr = 0
      Else
        rmndr = 5
      End If
    ElseIf (rmndr = 3) Then
      If prob <= (2# / 5#) Then
        rmndr = 0
      Else
        rmndr = 5
      End If
    ElseIf (rmndr = 4) Then
      If prob <= (1# / 5#) Then
        rmndr = 0
      Else
        rmndr = 5
      End If
    End If
    
    temp_table(x, y, 1) = temp_table(x, y, 0) - _
      (temp_table(x, y, 0) Mod 5) + rmndr

    'Print #temp, "Rndm rounding (base 5)", temp_table(x, y, 0), _
      temp_table(x, y, 0) Mod 5, prob, rmndr, temp_table(x, y, 1)

    ' Random rounding to nearest 3

    rmndr = temp_table(x, y, 0) Mod 3
    
    prob = Rnd
    If rmndr = 1 Then
      If prob <= (2# / 3#) Then
        rmndr = 0
      Else
        rmndr = 3
      End If
    ElseIf (rmndr = 2) Then
      If prob <= (2# / 3#) Then
        rmndr = 3
      Else
        rmndr = 0
      End If
    End If
    
    temp_table(x, y, 2) = temp_table(x, y, 0) - _
      (temp_table(x, y, 0) Mod 3) + rmndr

'    Print #temp, "Rndm rounding (base 3)", temp_table(x, y, 0), _
      temp_table(x, y, 2) Mod 3, prob, rmndr, temp_table(x, y, 2)

    
    ' Small number rounding [v3 - margins = sum of internal cells;
    '                        v6 = margins independently rounded]
    '(Here both versions have independently rounded margins; sub-routine "Recalc Marginals"
    ' recalcs margins as sum of internal cell counts for v3)
    
    If (temp_table(x, y, 0) < 3) Then
    
      rmndr = temp_table(x, y, 0) Mod 3
    
      prob = Rnd
      If rmndr = 1 Then
        If prob <= (2# / 3#) Then
          rmndr = 0
        Else
          rmndr = 3
        End If
      ElseIf (rmndr = 2) Then
        If prob <= (2# / 3#) Then
          rmndr = 3
        Else
          rmndr = 0
        End If
      End If
    
      temp_table(x, y, 3) = temp_table(x, y, 0) - _
        (temp_table(x, y, 0) Mod 3) + rmndr

    Else
    
      temp_table(x, y, 3) = temp_table(x, y, 0)
    
    End If

    temp_table(x, y, 6) = temp_table(x, y, 3)

'    Print #temp, "Small nos rndg", temp_table(x, y, 0), _
      temp_table(x, y, 0) Mod 3, prob, rmndr, temp_table(x, y, 3)

    
    'Barnardisation (p=0.1)

    rmndr = 0

    If (temp_table(x, y, 0) > 0) Then 'if not true zero
    
      prob1 = Rnd
      prob2 = Rnd
    
      If (prob1 <= 0.1) Then 'randomly add +/- 1 or 0
        If (prob2 <= 0.5) Then rmndr = -1 Else rmndr = 1
      Else 'leave unperturbed
        rmndr = 0
      End If
      
    End If 'if not true zero
    
    temp_table(x, y, 4) = temp_table(x, y, 0) + rmndr
    
'    Print #temp, "Barnardisation (p=0.1)", temp_table(x, y, 0), f(prob1, 8, 2), f(prob2, 8, 2), rmndr, temp_table(x, y, 4)

    'Barnardisation (p=0.02)

    rmndr = 0

    If (temp_table(x, y, 0) > 0) Then 'if not true zero
    
      prob1 = Rnd
      prob2 = Rnd
    
      If (prob1 <= 0.02) Then 'randomly add +/- 1 or 0
        If (prob2 <= 0.5) Then rmndr = -1 Else rmndr = 1
      Else 'leave unperturbed
        rmndr = 0
      End If
      
    End If 'if not true zero
    
    temp_table(x, y, 5) = temp_table(x, y, 0) + rmndr
    
'    Print #temp, "Barnardisation (p=0.02)", temp_table(x, y, 0), f(prob1, 8, 2), f(prob2, 8, 2), rmndr, temp_table(x, y, 5)

  Next y
Next x

'Write results to temporary file
'Print #temp, zoneid
'Call write_table(temp, 0, rows, cols, temp_table, tablename)
'Call write_table(temp, 1, rows, cols, temp_table, tablename)
'Call write_table(temp, 2, rows, cols, temp_table, tablename)
'Call write_table(temp, 3, rows, cols, temp_table, tablename)
'Call write_table(temp, 4, rows, cols, temp_table, tablename)
'Call write_table(temp, 5, rows, cols, temp_table, tablename)

End Sub
Sub Recalc_marginals(temp_table, map, rows, cols, temp, tablename)
Dim t, mapping, total3, total4, total5, marginal, x, y As Integer

'For small number rounding and barnardisation, marginals adjusted to agree with new numbers

'print #temp,
'print #temp, "Table  after mapping "; mapping
'print #temp, "Original data"
'For y = 1 To rows
'For x = 1 To cols
'print #temp, f(temp_table(x, y, 0), 5, 0);
'Next x
'print #temp,
'Next y

'print #temp,
'print #temp, "Method 2: random rounding (base 5)"
'For y = 1 To rows
'For x = 1 To cols
'print #temp, f(temp_table(x, y, 1), 5, 0);
'Next x
'print #temp,
'Next y

'print #temp,
'print #temp, "Method 3: random rounding (base 3)"
'For y = 1 To rows
'For x = 1 To cols
'print #temp, f(temp_table(x, y, 2), 5, 0);
'Next x
'print #temp,
'Next y

'print #temp,
'print #temp, "Method 4: small no. rounding"
'For y = 1 To rows
'For x = 1 To cols
'print #temp, f(temp_table(x, y, 3), 5, 0);
'Next x
'print #temp,
'Next y

'print #temp,
'print #temp, "Method 5: barndardisation (p=0.1)"
'For y = 1 To rows
'For x = 1 To cols
'print #temp, f(temp_table(x, y, 4), 5, 0);
'Next x
'print #temp,
'Next y

'print #temp,
'print #temp, "Method 6: barndardisation (p=0.02)"
'For y = 1 To rows
'For x = 1 To cols
'print #temp, f(temp_table(x, y, 5), 5, 0);
'Next x
'print #temp,
'Next y


For mapping = 1 To NoOfMaps

If (map(mapping, 0) = 2) Then 'row marginal
  For x = 1 To cols
    total3 = 0
    total4 = 0
    total5 = 0
    marginal = 0
    For y = 1 To rows
    If (map(mapping, y) < 0) Then marginal = y
      If (map(mapping, y) > 0) Then
        total3 = total3 + temp_table(x, y, 3)
        total4 = total4 + temp_table(x, y, 4)
        total5 = total5 + temp_table(x, y, 5)
      End If
    Next y 'next row
    temp_table(x, marginal, 3) = total3
    temp_table(x, marginal, 4) = total4
    temp_table(x, marginal, 5) = total5
  Next x 'next col
End If 'if row marginal

If (map(mapping, 0) = 1) Then 'col marginal
  For y = 1 To rows
    total3 = 0
    total4 = 0
    total5 = 0
    marginal = 0
    For x = 1 To cols
      If (map(mapping, x) < 0) Then marginal = x
      If (map(mapping, x) > 0) Then
      total3 = total3 + temp_table(x, y, 3)
      total4 = total4 + temp_table(x, y, 4)
      total5 = total5 + temp_table(x, y, 5)
    End If
    Next x 'next col
    temp_table(marginal, y, 3) = total3
    temp_table(marginal, y, 4) = total4
    temp_table(marginal, y, 5) = total5
  Next y 'next col
End If 'if row marginal

'print #temp,
'print #temp, "Table  after mapping "; mapping
'For y = 1 To rows
'For x = 1 To cols
'print #temp, f(temp_table(x, y, 3), 5, 0);
'Next x
'print #temp,
'Next y

'print #temp,
'print #temp, "Table after mapping "; mapping
'For y = 1 To rows
'For x = 1 To cols
'print #temp, f(temp_table(x, y, 4), 5, 0);
'Next x
'print #temp,
'Next y

'print #temp,
'print #temp, "Table after mapping "; mapping
'For y = 1 To rows
'For x = 1 To cols
'print #temp, f(temp_table(x, y, 5), 5, 0);
'Next x
'print #temp,
'Next y


Next mapping


'Write out table after recalculation of marginals
'Call write_table(temp, 0, rows, cols, temp_table, tablename)
'Call write_table(temp, 1, rows, cols, temp_table, tablename)
'Call write_table(temp, 2, rows, cols, temp_table, tablename)
'Call write_table(temp, 3, rows, cols, temp_table, tablename)
'Call write_table(temp, 4, rows, cols, temp_table, tablename)
'Call write_table(temp, 5, rows, cols, temp_table, tablename)

End Sub
Sub Recalc_raw_marginals(temp_table, map, rows, cols, temp)
Dim mapping, total0, marginal, x, y As Integer

'For raw SAS table, adjust marginals to agree with internal cell counts

' === Report raw marginals before adjustment ===
'Print #temp,
'Print #temp, "Raw Table before adjustment'"
'For y = 1 To rows
'For x = 1 To cols
'Print #temp, f(temp_table(x, y, 1), 5, 0);
'Next x
'Print #temp,
'Next y


' === Undertake adjustment ===
For mapping = 1 To NoOfMaps

If (map(mapping, 0) = 2) Then 'row marginal
  For x = 1 To cols
    total0 = 0
    marginal = 0
    For y = 1 To rows
    If (map(mapping, y) < 0) Then marginal = y
      If (map(mapping, y) > 0) Then total0 = total0 + temp_table(x, y, 0)
    Next y 'next row
    temp_table(x, marginal, 0) = total0
  Next x 'next col
End If 'if row marginal

If (map(mapping, 0) = 1) Then 'col marginal
  For y = 1 To rows
    total0 = 0
    marginal = 0
    For x = 1 To cols
      If (map(mapping, x) < 0) Then marginal = x
      If (map(mapping, x) > 0) Then total0 = total0 + temp_table(x, y, 0)
    Next x 'next col
    temp_table(marginal, y, 0) = total0
  Next y 'next col
End If 'if row marginal

Next mapping

' === Report raw table after adjustment ===
'Print #temp,
'Print #temp, "Raw Table after adjustment'"
'For y = 1 To rows
'For x = 1 To cols
'Print #temp, f(temp_table(x, y, 0), 5, 0);
'Next x
'Print #temp,
'Next y

End Sub
Sub close_table_specific_files(raw_input, mapfile, thresh_out, _
  perturb_1, perturb_2, perturb_3, perturb_4, perturb_5, perturb_6)

Close #raw_input
Close #mapfile
Close #thresh_out
Close #perturb_1
Close #perturb_2
Close #perturb_3
Close #perturb_4
Close #perturb_5
Close #perturb_6

End Sub

